% CLEANEDGELIST - remove short edges from a set of edgelists
%
% Function to clean up a set of edge lists generated by EDGELINK so that
% isolated edges and spurs that are shorter that a minimum length are removed.
% This code can also be use with a set of line segments generated by LINESEG.
%
% Usage: nedgelist = cleanedgelist(edgelist, minlength)
%
% Arguments:
%                edgelist - a cell array of edge lists in row,column coords in
%                           the form
%                           { [r1 c1   [r1 c1   etc }
%                              r2 c2    ...
%                                       ...
%                              rN cN]   ....]   
%                minlength - minimum edge length of interest
%
% Returns:
%                nedgelist - the new, cleaned up set of edgelists
%
% See also:  EDGELINK, DRAWEDGELIST, LINESEG

% Copyright (c) 2006 Peter Kovesi
% School of Computer Science & Software Engineering
% The University of Western Australia
% pk at csse uwa edu au
% http://www.csse.uwa.edu.au/
% 
% Permission is hereby granted, free of charge, to any person obtaining a copy
% of this software and associated documentation files (the "Software"), to deal
% in the Software without restriction, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included in 
% all copies or substantial portions of the Software.
%
% The Software is provided "as is", without warranty of any kind.
%
% December 2006

function nedgelist = cleanedgelist(edgelist, minlength)
    
    Nedges = length(edgelist);
    Nnodes = 2*Nedges;

    % Each edgelist has two end nodes - the starting point and the ending point.
    % We build up an adjacency/connection matrix for each node so that we can
    % determine which, if any, edgelists are connected to a node. We also
    % maintain an adjacency matrix for the edges themselves.
    % 
    % It is tricky maintaining all this information but it does allow the
    % code to run much faster.

    % First extract the end nodes from each edgelist.  The nodes are numbered
    % so that the start node has number 2*edgenumber-1 and the end node has
    % number 2*edgenumber
    node = zeros(Nnodes, 2);
    for n = 1:Nedges
	node(2*n-1,:) = edgelist{n}(1,:);
	node(2*n  ,:) = edgelist{n}(end,:);	
    end
    
    % Now build the adjacency/connection matrices. 
    A = zeros(Nnodes);   % Adjacency matrix for nodes
    B = zeros(Nedges);   % Adjacency matrix for edges
    
    for n = 1:Nnodes-1
	for m = n+1:Nnodes
            % If nodes m & n are connected
	    A(n,m) =  node(n,1)==node(m,1) && node(n,2)==node(m,2);
	    A(m,n) = A(n,m);
	    
	    if A(n,m)
		edgen = fix((n+1)/2);
		edgem = fix((m+1)/2);		
		B(edgen, edgem) = 1;
		B(edgem, edgen) = 1;		
	    end
	end
    end

    % If we sum the columns of the adjacency matrix we get the number of
    % other edgelists that are connected to an edgelist
    Nconnections = sum(A);   % Connection count array for nodes
    Econnections = sum(B);   % Connection count array for edges
    
    for n = 1:Nedges
	% If the edgelength is  below the minimum length
	if ~isempty(edgelist{n}) && edgelistlength(edgelist{n}) < minlength  
	    
	    % Remove unconnected lists
	    if ~Econnections(n) 
		removeedge(n);
	    
	    % Look for spurs coming from 4-way junctions that are below the minimum length
	    elseif isspur(n) == 3
		fprintf('4 way junction');
		removeedge(n);		

	    % Process edges that are spurs coming from a 3-way junction.
	    elseif isspur(n) == 2
		
		% Check the other edges coming from this point.  If any are
		% also spurs make sure we remove the shortest one
		linkingedges = find(B(n,:));
		spurs = n;
		len = edgelistlength(edgelist{n});
		for i = 1:length(linkingedges)
		    if isspur(linkingedges(i))
			spurs = [spurs linkingedges(i)];  
			len = [len edgelistlength(edgelist{linkingedges(i)})];  
		    end
		end

		linkingedges = [linkingedges n];
		
		[mn,i] = min(len);
		edge2delete = spurs(i);
		edges2merge = linkingedges(linkingedges ~= edge2delete);
		
		removeedge(edge2delete);		
		mergelists(edges2merge);				
	    end	
	end
    end
    
    % Final cleanup of any new isolated edges that might have been created by
    % removing spurs...
    
    for n = 1:Nedges
	if ~isempty(edgelist{n}) && edgelistlength(edgelist{n}) < minlength  
	    if ~Econnections(n) 
		removeedge(n);		
	    end
	end
    end
    
    % Run through the edgelist and extract out the non-empty lists
    m = 0;
    for n = 1:Nedges
       if ~isempty(edgelist{n})
	   m = m+1;
	   nedgelist{m} = edgelist{n};
       end
    end
	
%----------------------------------------------------------------------    
% Internal function to merge 2 edgelists together and perform the necessary
% updates to the edge adjacency and node adjacency matrices and the connection
% count arrays

function mergelists(indices)

    if length(indices) ~= 2
	error('Attempt to merge other than 2 lists');
    end
    
    edge1 = indices(1);
    edge2 = indices(2);

    % Get indices of nodes at each end of the two edges
    s1 = 2*edge1-1; e1 = 2*edge1;
    s2 = 2*edge2-1; e2 = 2*edge2;    
    
    % Join edgelists together - with appropriate reordering depending on which
    % end is connected to which.  The result is stored in edge1
    
    if A(e1,s2)
	edgelist{edge1} = [edgelist{edge1}; edgelist{edge2}];

	A(e1,:) = A(e2,:); 	A(:,e1) = A(:,e2);
	Nconnections(e1) = Nconnections(e2);
	
    elseif  A(e1,e2)
	edgelist{edge1} = [edgelist{edge1}; flipud(edgelist{edge2})];	
	
	A(e1,:) = A(s2,:); 	A(:,e1) = A(:,s2);
	Nconnections(e1) = Nconnections(s2);

    elseif  A(s1,s2)
	edgelist{edge1} = [flipud(edgelist{edge1}); edgelist{edge2}]; 
	
	A(s1,:) = A(e1,:); 	A(:,s1) = A(:,e1);
	A(e1,:) = A(e2,:); 	A(:,e1) = A(:,e2);	
	Nconnections(s1) = Nconnections(e1);
	Nconnections(e1) = Nconnections(e2);	

    elseif  A(s1,e2)
	edgelist{edge1} = [flipud(edgelist{edge1}); flipud(edgelist{edge2})];

	A(s1,:) = A(e1,:); 	A(:,s1) = A(:,e1);	    
	A(e1,:) = A(s2,:); 	A(:,e1) = A(:,s2);
	Nconnections(s1) = Nconnections(e1);
	Nconnections(e1) = Nconnections(s2);
	
    else
	error('We should not have got here - edgelists cannot be merged');
    end

    
    % Now correct the edge adjacency matrix to reflect the new arrangement
    % The edges that the new edge1 is connected to is all the edges that
    % edge1 and edge2 were connected to
    B(edge1,:) = B(edge1,:) | B(edge2,:);
    B(:,edge1) = B(:,edge1) | B(:,edge2);    
    B(edge1, edge1) = 0;

    % Recompute connection counts because we have shuffled the adjacency matrices
    Econnections = sum(B);
    Nconnections = sum(A);
  
    removeedge(edge2);  % Finally discard edge2
    
    end

%--------------------------------------------------------------------    
% Function to test whether an edgelist is a spur or not and if so return the
% number of other edgelists it is connected to
    
function t = isspur(n)
    sconns = Nconnections(2*n-1);  % No of connections to start node
    econns = Nconnections(2*n);	   % No of connections to end node    
    
    t = (sconns == 0 && econns == 2) || (sconns == 2 && econns == 0);
    t = t*max(sconns,econns);
end


%--------------------------------------------------------------------
% Function to remove an edgelist and perform the necessary updates to the edge
% adjacency and node adjacency matrices and the connection count arrays

function removeedge(n)
    
    edgelist{n} = [];
    Econnections = Econnections - B(n,:); 
    Econnections(n) = 0; 
    B(n,:) = 0;
    B(:,n) = 0;
    
    nodes2delete = [2*n-1, 2*n];
    
    Nconnections = Nconnections - A(nodes2delete(1),:);    
    Nconnections = Nconnections - A(nodes2delete(2),:);        
    
    A(nodes2delete, :) = 0;
    A(:, nodes2delete) = 0;		    
    
end

%--------------------------------------------------------------------
% Function to compute the path length of an edgelist

function l = edgelistlength(edgelist)
    l = sum(sqrt(sum((edgelist(1:end-1,:)-edgelist(2:end,:)).^2, 2)));
end

%--------------------------------------------------------------------
end % End of cleanedgelists
    
